/*
 * Created by Angel Leon (@gubatron), Alden Torres (aldenml)
 * Copyright (c) 2011-2014, FrostWire(R). All rights reserved.
 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.limegroup.gnutella.gui;

import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Map;

import javax.swing.AbstractButton;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.plaf.ComponentUI;
import javax.swing.text.BadLocationException;

import net.miginfocom.swing.MigLayout;

import com.frostwire.gui.library.LibraryMediator;
import com.frostwire.gui.player.MediaPlayerComponent;
import com.frostwire.gui.searchfield.GoogleSearchField;
import com.frostwire.gui.searchfield.SearchField;
import com.frostwire.gui.tabs.LibraryTab;
import com.frostwire.gui.tabs.Tab;
import com.frostwire.gui.theme.SkinApplicationHeaderUI;
import com.frostwire.gui.theme.ThemeMediator;
import com.frostwire.gui.updates.UpdateMediator;
import com.frostwire.uxstats.UXAction;
import com.frostwire.uxstats.UXStats;
import com.limegroup.gnutella.MediaType;
import com.limegroup.gnutella.gui.GUIMediator.Tabs;
import com.limegroup.gnutella.gui.actions.FileMenuActions;
import com.limegroup.gnutella.gui.search.SearchInformation;
import com.limegroup.gnutella.gui.search.SearchMediator;
import com.limegroup.gnutella.settings.SearchSettings;
import com.limegroup.gnutella.util.URLDecoder;

/**
 * 
 * @author gubatron
 * @author aldenml
 *
 */
public final class ApplicationHeader extends JPanel implements RefreshListener {

    /*
    * The property to store the selected icon in.
    */
    private static final String SELECTED = "SELECTED_ICON";

    /**
     * The property to store the unselected icon in.
     */
    private static final String DESELECTED = "DESELECTED_ICON";

    private static final String CLOUD_SEARCH_FIELD = "cloud_search_field";

    private static final String LIBRARY_SEARCH_FIELD = "library_search_field";

    /**
     * The clicker forwarder.
     */
    private final MouseListener CLICK_FORWARDER = new Clicker();

    /** Button background for selected button */
    private final Image headerButtonBackgroundSelected;

    /** Button background for unselected button */
    private final Image headerButtonBackgroundUnselected;

    private LogoPanel logoPanel;
    
    private ImageIcon chatButtonImage;
    private JRadioButton chatButton;

    private JLabel updateButton;
    private ImageIcon updateImageButtonOn;
    private ImageIcon updateImageButtonOff;
    private long updateButtonAnimationStartedTimestamp;

    private final GoogleSearchField cloudSearchField;
    private SearchField librarySearchField;
    private final JPanel searchPanels;

    public ApplicationHeader(Map<Tabs, Tab> tabs) {
        setMinimumSize(new Dimension(1000, 54));
        setLayout(new MigLayout("ins 0, ay 50%, filly,", "[][][grow][][]"));

        headerButtonBackgroundSelected = GUIMediator.getThemeImage("selected_header_button_background").getImage();
        headerButtonBackgroundUnselected = GUIMediator.getThemeImage("unselected_header_button_background").getImage();

        cloudSearchField = new GoogleSearchField() {
            {
                super.getFindButton().addActionListener(new ActionListener() {
                    @Override
                    public void actionPerformed(ActionEvent e) {
                        UXStats.instance().log(UXAction.SEARCH_STARTED_SMALL_SEARCH_ICON_CLICK);
                    }
                    
                });
            }
            
            @Override
            public void processKeyEvent(KeyEvent evt) {
                if ((evt.getID()  == KeyEvent.KEY_PRESSED) && 
                    evt.getKeyCode() == KeyEvent.VK_ENTER) {
                    UXStats.instance().log(UXAction.SEARCH_STARTED_ENTER_KEY);
                }
                super.processKeyEvent(evt);
            }
        };
        
        searchPanels = createSearchPanel();
        add(searchPanels, "w 240px!, gapright 10px, gapleft 5px");

        //The Chat Tab is not a real Tab, it's a button, it's creation is done inside this method.
        addTabButtons(tabs);

        createUpdateButton();
        JPanel logoUpdateButtonsPanel = new JPanel();
        logoPanel = new LogoPanel();
        //only one will be shown at the time.
        logoUpdateButtonsPanel.add(logoPanel);
        logoUpdateButtonsPanel.add(updateButton);

        add(logoUpdateButtonsPanel, "growx, alignx center");

        JComponent player = new MediaPlayerComponent().getMediaPanel();
        add(player, "dock east, growy, gapafter 10px!");

        GUIMediator.addRefreshListener(this);

        final ActionListener schemaListener = new SchemaListener();
        schemaListener.actionPerformed(null);
    }


    private JPanel createSearchPanel() {
        JPanel panel = new JPanel(new CardLayout());
        initCloudSearchField();
        createLibrarySearchField();

        panel.add(cloudSearchField, CLOUD_SEARCH_FIELD);
        panel.add(librarySearchField, LIBRARY_SEARCH_FIELD);

        return panel;
    }

    private void createLibrarySearchField() {
        librarySearchField = LibraryMediator.instance().getLibrarySearch().getSearchField();
    }

    private void initCloudSearchField() {
        cloudSearchField.addActionListener(new SearchListener());
        cloudSearchField.setPrompt(I18n.tr("Search or enter URL"));
        Font origFont = cloudSearchField.getFont();
        Font newFont = origFont.deriveFont(origFont.getSize2D() + 2f);
        cloudSearchField.setFont(newFont);
        cloudSearchField.setMargin(new Insets(0, 2, 0, 0));
    }

    private void createUpdateButton() {
        updateImageButtonOn = GUIMediator.getThemeImage("update_button_on");
        updateImageButtonOff = GUIMediator.getThemeImage("update_button_off");

        updateButton = new JLabel(updateImageButtonOn);
        updateButton.setToolTipText(I18n.tr("A new update has been downloaded."));
        Dimension d = new Dimension(32, 32);

        updateButton.setVisible(false);
        updateButton.setSize(d);
        updateButton.setPreferredSize(d);
        updateButton.setMinimumSize(d);
        updateButton.setMaximumSize(d);
        updateButton.setBorder(null);
        updateButton.setOpaque(false);

        updateButtonAnimationStartedTimestamp = -1;

        updateButton.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                UpdateMediator.instance().showUpdateMessage();
            }
        });
    }
    
    private void createChatButton() {
        chatButtonImage = GUIMediator.getThemeImage("chat_tab");


        chatButton = new JRadioButton(I18n.tr("Chat"));
        Dimension d = new Dimension(65, 55);
        Font buttonFont = new Font("Helvetica", Font.BOLD, 10);
        
        chatButton.setFont(buttonFont);
        chatButton.setIcon(chatButtonImage);
        chatButton.setPressedIcon(chatButtonImage);
        chatButton.setSelectedIcon(chatButtonImage);
        chatButton.setRolloverIcon(chatButtonImage);
        chatButton.setRolloverSelectedIcon(chatButtonImage);
        chatButton.setVisible(true);
        chatButton.setSize(d);
        chatButton.setPreferredSize(d);
        chatButton.setMinimumSize(d);
        chatButton.setMaximumSize(d);
        chatButton.setBorder(null);
        chatButton.setBorderPainted(false);
        chatButton.setFocusPainted(false);
        chatButton.setContentAreaFilled(false);
        chatButton.setOpaque(false);
        chatButton.setToolTipText(I18n.tr("Show our community chat"));
        chatButton.setHorizontalTextPosition(SwingConstants.CENTER);
        chatButton.setVerticalTextPosition(JLabel.BOTTOM);
        chatButton.setHorizontalAlignment(SwingConstants.CENTER);
        chatButton.setForeground(ThemeMediator.TAB_BUTTON_FOREGROUND_COLOR);

        chatButton.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                GUIMediator.openURL("http://www.frostwire.com/chat");
                UXStats.instance().log(UXAction.MISC_CHAT_OPENED_IN_BROWSER);
            }
        });
        
    }


    public LogoPanel getLogoPanel() {
        return logoPanel;
    }

    private void addTabButtons(final Map<Tabs, Tab> tabs) {
        JPanel buttonContainer = new JPanel(new MigLayout("insets 0, gap 0"));
        ButtonGroup group = new ButtonGroup();

        Font buttonFont = new Font("Helvetica", Font.BOLD, 10);
        buttonContainer.add(ThemeMediator.createAppHeaderSeparator(), "growy");

        for (Tabs t : GUIMediator.Tabs.values()) {
            final Tabs lameFinalT = t; //java...
            AbstractButton button = createTabButton(tabs.get(t));
            button.setFont(buttonFont);

            button.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent e) {
                    onMainApplicationHeaderTabActionPerformed(tabs, lameFinalT);
                }

                private void onMainApplicationHeaderTabActionPerformed(final Map<Tabs, Tab> tabs, final Tabs lameFinalT) {
                    prepareSearchTabAsSearchTrigger(lameFinalT);
                    showSearchField(tabs.get(lameFinalT));
                    GUIMediator.instance().setWindow(lameFinalT);
                }
                
                public void prepareSearchTabAsSearchTrigger(final Tabs lameFinalT) {
                    boolean performInternetSearch = false;
                    String query = null;
                    if (lameFinalT == Tabs.SEARCH) {
                        if (!cloudSearchField.getText().isEmpty()) {
                            performInternetSearch = true;
                            query = cloudSearchField.getText();
                        } else if (cloudSearchField.getText().isEmpty() && !librarySearchField.getText().isEmpty()) {
                            //they want internet search while on the library
                            performInternetSearch = true;
                            query = librarySearchField.getText();
                            librarySearchField.setText(new String(""));
                            cloudSearchField.setText(query);
                        }

                        if (query != null) {
                            if (performInternetSearch) {
                                UXStats.instance().log(UXAction.SEARCH_STARTED_SEARCH_TAB_BUTTON);
                                cloudSearchField.getActionListeners()[0].actionPerformed(null);
                            }
                        }
                    }
                }
            });

            group.add(button);
            buttonContainer.add(button);
            buttonContainer.add(ThemeMediator.createAppHeaderSeparator(), "growy, w 0px");

            button.setSelected(t.equals(GUIMediator.Tabs.SEARCH));
        }
        
        //CHAT BUTTON
        createChatButton();
        buttonContainer.add(chatButton);
        buttonContainer.add(ThemeMediator.createAppHeaderSeparator(), "growy, w 0px");

        add(buttonContainer, "");
    }
    
    
 
    /** Given a Tab mark that button as selected 
     * 
     * Since we don't keep explicit references to the buttons this method
     * walks over the components in the ApplicationHeader until it finds
     * the AbstractButton that has the Tab object as a client property named "tab" 
     * @see MainFrame#setSelectedTab(Tabs)
     */
    public void selectTab(Tab t) {
        Component[] components = getComponents();
        JPanel buttonContainer = (JPanel) components[1];
        Component[] buttons = buttonContainer.getComponents();
        for (Component c : buttons) {
            if (c instanceof AbstractButton) {
                AbstractButton b = (AbstractButton) c;
                if (b.getClientProperty("tab").equals(t)) {
                    b.setSelected(true);
                    return;
                }
            }
        }
    }

    private AbstractButton createTabButton(Tab t) {
        Icon icon = t.getIcon();
        Icon disabledIcon = null;
        Icon rolloverIcon = null;

        final JRadioButton button = new JRadioButton(I18n.tr(t.getTitle())) {
            protected void paintComponent(Graphics g) {
                if (isSelected()) {
                    g.drawImage(headerButtonBackgroundSelected, 0, 0, null);
                } else {
                    g.drawImage(headerButtonBackgroundUnselected, 0, 0, null);
                }
                super.paintComponent(g);
            }
        };

        button.putClientProperty("tab", t);

        button.putClientProperty(SELECTED, icon);
        if (icon != null) {
            disabledIcon = icon;//ImageManipulator.darken(icon);
            rolloverIcon = icon;//ImageManipulator.brighten(icon);
        }
        button.putClientProperty(DESELECTED, disabledIcon);
        button.setIcon(disabledIcon);
        button.setPressedIcon(rolloverIcon);
        button.setSelectedIcon(disabledIcon);
        button.setRolloverIcon(rolloverIcon);
        button.setRolloverSelectedIcon(rolloverIcon);
        button.setBorderPainted(false);
        button.setFocusPainted(false);
        button.setContentAreaFilled(false);
        button.setOpaque(false);
        button.addMouseListener(CLICK_FORWARDER);
        button.setToolTipText(t.getToolTip());
        button.setHorizontalTextPosition(SwingConstants.CENTER);
        button.setVerticalTextPosition(SwingConstants.BOTTOM);
        button.setHorizontalAlignment(SwingConstants.CENTER);
        button.setForeground(ThemeMediator.TAB_BUTTON_FOREGROUND_COLOR);

        Dimension buttonDim = new Dimension(65, 55);
        button.setPreferredSize(buttonDim);
        button.setMinimumSize(buttonDim);
        button.setMaximumSize(buttonDim);
        button.setSelected(false);

        return button;
    }

    public void showSearchField(Tab t) {
        cloudSearchField.setText("");
        librarySearchField.setText("");
        CardLayout cl = (CardLayout) (searchPanels.getLayout());
        if (t instanceof LibraryTab) {
            cl.show(searchPanels, LIBRARY_SEARCH_FIELD);
        } else {
            cl.show(searchPanels, CLOUD_SEARCH_FIELD);
        }
    }

    /**
     * Forwards click events from a panel to the panel's component.
     */
    private static class Clicker extends MouseAdapter {
        public void mouseClicked(MouseEvent e) {
            JComponent c = (JComponent) e.getSource();
            if (!(c instanceof AbstractButton)) {
                AbstractButton b = (AbstractButton) c.getComponent(0);
                b.doClick();
            }
        }
    }

    @Override
    public void refresh() {
        showUpdateButton(!UpdateMediator.instance().isUpdated() && UpdateMediator.instance().isUpdateDownloaded());
    }

    public static class IntermittentButton {
        private JLabel buttonReference;
        private ImageIcon imgOn;
        private ImageIcon imgOff;

        /** Pass the label you want to animate */
        public IntermittentButton(JLabel button, ImageIcon on, ImageIcon off) {
            buttonReference = button;
            imgOn = on;
            imgOff = off;
        }

        public void setImage(final boolean on) {
            GUIMediator.safeInvokeLater(new Runnable() {
                @Override
                public void run() {
                    buttonReference.setIcon((on) ? imgOn : imgOff);
                }
            });
        }
    }

    private void showUpdateButton(boolean show) {
        if (updateButton.isVisible() == show) {
            return;
        }

        logoPanel.setVisible(!show);
        updateButton.setVisible(show);

        if (show) {
            //Start animating the button for 30 seconds.
            if (updateButtonAnimationStartedTimestamp == -1) {
                startUpdateButtonIntermittentAnimation();
            }
        }
    }

    private void startUpdateButtonIntermittentAnimation() {
        updateButtonAnimationStartedTimestamp = System.currentTimeMillis();

        //start animation thread.
        Thread t = new Thread("update-button-animation") {
            private final long ANIMATION_DURATION = 30000;
            private final long ANIMATION_INTERVAL = 1000;
            private long updateButtonAnimationLastChange;

            public void run() {
                long now = System.currentTimeMillis();
                updateButtonAnimationLastChange = now;

                boolean buttonState = true;
                while (now - updateButtonAnimationStartedTimestamp < ANIMATION_DURATION) {
                    if (now - updateButtonAnimationLastChange >= ANIMATION_INTERVAL) {
                        switchButtonImage(buttonState);
                        buttonState = !buttonState;
                    }
                    try {
                        sleep(ANIMATION_INTERVAL);
                    } catch (InterruptedException e) {
                    }
                    now = System.currentTimeMillis();
                }
                switchButtonImage(false);
            }

            public void switchButtonImage(final boolean state) {
                updateButtonAnimationLastChange = System.currentTimeMillis();
                GUIMediator.safeInvokeLater(new Runnable() {
                    public void run() {
                        updateButton.setIcon(state ? updateImageButtonOn : updateImageButtonOff);
                    }
                });
            }
        };
        t.start();
    }

    @Override
    public void updateUI() {
        ComponentUI ui = UIManager.getUI(this);
        if (ui == null) {
            ui = new SkinApplicationHeaderUI();
        }
        setUI(ui);
    }

    @Override
    public String getUIClassID() {
        return "ApplicationHeaderUI";
    }

    public void requestSearchFocusImmediately() {
        if (cloudSearchField != null) {
            cloudSearchField.requestFocus();
        }
    }

    public void requestSearchFocus() {
        // Workaround for bug manifested on Java 1.3 where FocusEvents
        // are improperly posted, causing BasicTabbedPaneUI to throw an
        // ArrayIndexOutOfBoundsException.
        // See:
        // http://developer.java.sun.com/developer/bugParade/bugs/4523606.html
        // http://developer.java.sun.com/developer/bugParade/bugs/4379600.html
        // http://developer.java.sun.com/developer/bugParade/bugs/4128120.html
        // for related problems.
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                requestSearchFocusImmediately();
            }
        });
    }
    
    public void startSearch(String query) {
        cloudSearchField.setText(query);
        cloudSearchField.getActionListeners()[0].actionPerformed(null);
    }

    private class SearchListener implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            //keep the query if there was one before switching to the search tab
            String query = cloudSearchField.getText();
            String queryTitle = query;
            GUIMediator.instance().setWindow(GUIMediator.Tabs.SEARCH);                

            //start a download from the search box by entering a URL.
            if (FileMenuActions.openMagnetOrTorrent(query,FileMenuActions.ActionInvocationSource.FROM_SEARCH_FIELD)) {
                cloudSearchField.setText("");
                cloudSearchField.hidePopup();
                return;
            }
            
            
            if (query.contains("www.frostclick.com/cloudplayer/?type=yt") ||
                query.contains("frostwire-preview.com/?type=yt")) {
                try {
                    query = query.split("detailsUrl=")[1];
                    query = URLDecoder.decode(query);
                } catch (Exception e1) {
                    e1.printStackTrace();
                }
            }
            
            if (query.contains("youtube.com/watch?")) {
                query = query.split("v=")[1];
                if (query.contains("#")) {
                    query = query.split("#")[0];
                }
                queryTitle = "youtube:" + query;
            }

            if (query.contains("http://youtu.be/")) {
                query = query.split("be/")[1];
                if (query.contains("#")) {
                    query = query.split("#")[0];
                }
                queryTitle = "youtube:" + query;
            }

            final SearchInformation info = SearchInformation.createTitledKeywordSearch(query, null, MediaType.getTorrentMediaType(), queryTitle);

            // If the search worked, store & clear it.
            if (SearchMediator.instance().triggerSearch(info) != 0) {
                if (info.isKeywordSearch()) {

                    cloudSearchField.addToDictionary();

                    // Clear the existing search.
                    cloudSearchField.setText("");
                    cloudSearchField.hidePopup();
                }
            }
        }
    }

    private class SchemaListener implements ActionListener {
        public void actionPerformed(ActionEvent event) {
            SearchSettings.MAX_QUERY_LENGTH.revertToDefault();

            //Truncate if you have too much text for a gnutella search
            if (cloudSearchField.getText().length() > SearchSettings.MAX_QUERY_LENGTH.getValue()) {
                try {
                    cloudSearchField.setText(cloudSearchField.getText(0, SearchSettings.MAX_QUERY_LENGTH.getValue()));
                } catch (BadLocationException e) {
                }
            }

            requestSearchFocus();
        }
    }
}
